using LinearAlgebra, Optim, LineSearches, JLD2, StellaratorOptimization, NE3DLE
#ENV["JULIA_DEBUG"] = StellOpt

BLAS.set_num_threads(4)
optimizer = StellOpt.Optimizer(GradientDescent(; alphaguess=1e+0, linesearch=Static()),
                               Optim.Options(iterations = 500, store_trace = true,
                                             extended_trace = true, f_tol = 1e-16, 
                                             g_tol = 1e-16, allow_f_increases = true))
prob = StellOpt.Problem()
prob.output_tag = "lha_nfp_2_run_1"
prob.output_directory = pwd()
StellOpt.add_optimizer!(prob, optimizer)

fd = StellOpt.CenteredDiff(1e-5)
spsa = StellOpt.SPSA(α = 0.602, β = 100_000, γ = 0.101, δ = 1e-5, ϵ = 1e-8, n = 32, max_evals = 1)
nml = NE3DLE.read_par_file("NE3DLE.in")
eq = call_NE3DLE(nml)

Ne3dle_wrapper = StellOpt.EquilibriumWrapper(nml, eq, spsa)
StellOpt.add_equilibrium_wrapper!(prob, Ne3dle_wrapper)
Ne3dle_s = StellOpt.@add_derived_geometry!(prob, NE3DLESurface(equilibrium=Ne3dle_wrapper))
StellOpt.@add_target!(prob, NE3DLE.quasisymmetry_deviation(Ne3dle_s), 0.0, 1e-2, "qs")
StellOpt.@add_target!(prob, NE3DLE.flux_surface_averaged_parallel_current(Ne3dle_s), 0.0, 1e+0, "<j_parallel>")
StellOpt.@add_target!(prob, NE3DLE.rotational_transform(Ne3dle_s), π^2/9, 1e+0, "iota")
StellOpt.@add_target!(prob, NE3DLE.global_shear(Ne3dle_s), 0.0, 1e+0, "shear")
StellOpt.@add_target!(prob, NE3DLE.average_aspect_ratio(Ne3dle_s), 10.0, 1e+0, "aspect")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(0,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(0,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(0,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(0,4)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(0,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(0,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(0,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(0,4)")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(1,-4)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(1,-3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(1,-2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(1,-1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(1,0)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(1,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(1,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(1,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(1,4)")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(1,-4)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(1,-3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(1,-2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(1,-1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(1,0)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(1,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(1,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(1,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(1,4)")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(2,-4)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(2,-3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(2,-2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(2,-1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(2,0)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(2,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(2,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(2,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(2,4)")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(2,-4)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(2,-3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(2,-2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(2,-1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(2,0)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(2,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(2,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(2,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(2,4)")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(3,-4)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(3,-3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(3,-2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(3,-1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(3,0)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(3,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(3,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(3,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(3,4)")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(3,-4)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(3,-3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(3,-2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(3,-1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(3,0)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(3,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(3,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(3,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(3,4)")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(4,-4)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(4,-3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(4,-2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(4,-1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(4,0)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(4,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(4,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(4,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Rmn(4,4)")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(4,-4)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(4,-3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(4,-2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(4,-1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(4,0)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(4,1)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(4,2)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(4,3)")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "Zmn(4,4)")

StellOpt.add_variable!(prob, Ne3dle_wrapper, "s_param")
StellOpt.add_variable!(prob, Ne3dle_wrapper, "iota")

residual = StellOpt.optimize_threads(prob)

JLD2.save_object(joinpath(@__DIR__, prob.output_tag*"_residual.jld"), residual)

foreach(t -> println(t.label, " => ", t.current_value), values(prob.targets))
